/* 
 * Implémentation des fonctions natives
 * pour les API java.lang.*
 */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "native.h"

#include "api_lang.h"


/* *************************************************************
 * 
 * JAVA.LANG.THREAD
 *
 */

Object *
java_lang_Thread_currentThread(arg_list args)
{
	USED(args);

	return getCurrentThread();
}

void
java_lang_Thread_yield(arg_list args)
{
	USED(args);
	
	yieldThread();
}

void
java_lang_Thread_sleep(arg_list args)
{
	vlong 	ms 		=  arg_value(args, vlong);
	
	sleepThread(ms);
}

void
java_lang_Thread_setPriority0(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	int prio 		=  arg_value(args, int);
	
	setPriorityThread(thread, prio);
}

void
java_lang_Thread_stop0(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	Object *th 		=  arg_value(args, Object *);
	USED(thread);

	stopThread(thread, th);
}

void
java_lang_Thread_start(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	
	startThread(thread);
}

void
java_lang_Thread_suspend0(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	
	suspendThread(thread);
}

void
java_lang_Thread_resume0(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	
	resumeThread(thread);
}

int
java_lang_Thread_isAlive(arg_list args)
{
	Object *thread 	=  arg_value(args, Object *);
	
	return isAliveThread(thread);
}

/* *************************************************************
 * 
 * JAVA.LANG.OBJECT
 *
 */

void 
java_lang_Object_wait(arg_list args)
{
	Object *object 	=  	arg_value(args, Object *);
	vlong timeout 	=  	arg_value(args, vlong);

	waitThread(object, timeout);
}

void 
java_lang_Object_notify(arg_list args)
{
	Object *object 	=  	arg_value(args, Object *);

	notifyThread(object);
}

void
java_lang_Object_notifyAll(arg_list args)
{
	Object *object =  arg_value(args, Object *);

	notifyAllThread(object);	
}

int 
java_lang_Object_hashCode(arg_list args)
{
	Object *object =  arg_value(args, Object *);
	
	return (int)object;
}

Object *
java_lang_Object_clone(arg_list args)
{
	Object *obj =  arg_value(args, Object *);
	
	Object *new_obj;

	if( IS_ARRAY(obj) || IS_ARRAY_OBJECT(obj) )
		panic("clone object tab not implemented\n");
	
	new_obj = newObject(obj->class);
	memcpy(new_obj->vars, obj->vars, obj->vars_size);

	return new_obj;
}

Object *
java_lang_Object_getClass(arg_list args)
{
	Object *obj =  arg_value(args, Object *);

	return obj->class->r_java_class;
}



/* *************************************************************
 * 
 * JAVA.LANG.CLASS
 *
 */

Object *
java_lang_Class_forName(arg_list args)
{
	Object *string = arg_value(args, Object *);

	ClassFile *class;
	char class_name[256], *pt;
	
	javaString2CString(string, class_name, sizeof class_name);
	for(pt=class_name; *pt ; pt++)
		if( *pt == '.' )
			*pt = '/';
	
	class = lookupClass(class_name);
	if( !class ){
		throwException(getCurrentFrame(), "java/lang/ClassNotFoundException");
		return 0;
	}

	return class->r_java_class;
}

Object *
java_lang_Class_newInstance(arg_list args)
{
	Object *obj = arg_value(args, Object *);

	API_JavaLangClass *o = U(API_JavaLangClass, obj);
	ClassFile *class = o->class_file;
	Object *new_object;

	/* creation de l'objet */
	new_object = newObject(class);
	
	executeConstructor(new_object, "()V");

	return new_object;
}

Object *
java_lang_Class_getName(arg_list args)
{
	Object *obj = arg_value(args, Object *);

	API_JavaLangClass *o = U(API_JavaLangClass, obj);
	ClassFile *class = o->class_file;
	Object *string;
	char name[256];
	int i;

	for(i=0 ; class->r_name[i] ; i++)
		if( class->r_name[i] == '/' )
			name[i] = '.';
		else
			name[i] = class->r_name[i];	
	name[i] = 0;

	string = makeJavaString(name);
	return string;
}

Object *
java_lang_Class_getSuperclass(arg_list args)
{
	Object *obj = arg_value(args, Object *);

	API_JavaLangClass *o = U(API_JavaLangClass, obj);

	return (o->class_file)->r_super->r_java_class;
}

Object *
java_lang_Class_getInterfaces(arg_list args)
{
	Object *obj = arg_value(args, Object *);

	API_JavaLangClass *o = U(API_JavaLangClass, obj);
	Object *array;
	ClassFile *c, *cc;
	int i;

	c = o->class_file;
	array = newObjectArray(obj->class, c->interfaces_count);
	
	for(i=0 ; i < c->interfaces_count ; i++){
		if( !c->r_resolved[c->interfaces[i]] )
			cp_resolution(c, c->interfaces[i], 0);
		if( getCurrentFrame()->exception )
			return 0;
		cc = (ClassFile *)c->r_resolved[c->interfaces[i]];
		array->vars[i] = (int)cc->r_java_class;
	}

	return array;
}

Object *
java_lang_Class_getClassLoader(arg_list args)
{
	USED(args);

	return 0;
}

int 
java_lang_Class_isInterface(arg_list args)
{
	Object *obj = arg_value(args, Object *);


	API_JavaLangClass *o = U(API_JavaLangClass, obj);

	return ((o->class_file)->access_flags & ACC_INTERFACE) ? 1 : 0;
}	

/* *************************************************************
 * 
 * JAVA.LANG.THROWABLE
 *
 */

Object *
java_lang_Throwable_fillInStackTrace(arg_list args)
{
	Object * thr = arg_value(args, Object *);

	fillInStackTrace(thr); /* fonction dans frame.c */

	return thr;
}

void
java_lang_Throwable_printStackTrace0(arg_list args)
{
	Object *thr 	= arg_value(args, Object *);
	Object *stream 	= arg_value(args, Object *);
	
	Object *backtrace = (Object *)thr->vars[0];
	Object *string;
	char buf[1024];
	int i, count;

	/* construire la chaine a afficher */
	for(i = count = 0 ; i < backtrace->vars_size/sizeof(int) ; i+=2){
		char name[256]; int u;

/*		chat("i = %d\n", i);
		chat("%s at %d\n", 
			((MethodBlock *)backtrace->vars[i])->name, 
			backtrace->vars[i+1]);
*/
		for(u=0 ; (name[u] = *(((MethodBlock *)backtrace->vars[i])->name + u)) != '(' ; u++);
		name[u] = 0;
		count += sprint(buf + count, "\tat %s.%s(%s:%d)\n", 
							((MethodBlock *)backtrace->vars[i])->class->r_name,
							name,
							((MethodBlock *)backtrace->vars[i])->class->r_file,
							backtrace->vars[i+1]);
	}
	
	/* creation de l'objet string */
	string = makeJavaString(buf);
	
	/* excution de println */
	executeVirtual(stream->class, stream, "print", "(Ljava/lang/String;)V", string);
}


/* *************************************************************
 * 
 * JAVA.LANG.SYSTEM
 *
 */

void
java_lang_System_arraycopy(arg_list args)
{
	Object *	src 			= arg_value(args, Object *);
	int 		src_position 	= arg_value(args, int);
	Object *	dst 			= arg_value(args, Object *);
	int 		dst_position 	= arg_value(args, int);
	int 		length 			= arg_value(args, int);
	
	int size;
	//print("%d %d %d %d %d\n", src, src_position, dst, dst_position, length);
	if( !src || !dst ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}	

	if(src->type == T_CHAR){
		size = 2;
	}else if(src->type == T_BYTE){
		size = 1;
	}else{
		size = typeSize(src->type) * sizeof(int);
	}

	if( src_position + length > arrayLength(src) ){
		throwException(getCurrentFrame(), "java/lang/ArrayIndexOutOfBoundsException");
		return;
	}
	if( dst_position + length > arrayLength(dst) ){
		throwException(getCurrentFrame(), "java/lang/ArrayStoreException");
		return;
	} 
	
	memmove((char *)dst->vars + dst_position * size, 
				(char *)src->vars + src_position * size, length * size);
}

vlong
java_lang_System_currentTimeMillis(void)
{
	vlong t0 = time(0);
	vlong t1 = getMiliSec();
	vlong t;

	t = t1 / 1000;
	t *= 1000;

	t0 = (t0 * 1000) - t + t1;

	return t0;
}

Object *
java_lang_System_initProperties(arg_list args)
{
	Object *	prop =	arg_value(args, Object *);
	
	char buf[512];
	extern char *java_version;	
	extern char *java_vendor;
	extern char *java_vendor_url;
	extern char *java_home;
	extern char *java_class_path;

	setProperty(prop, "java.version", 		java_version);
	setProperty(prop, "java.vendor", 		java_vendor);
	setProperty(prop, "java.vendor.url", 	java_vendor_url);
	setProperty(prop, "java.home", 			java_home);
	setProperty(prop, "java.class.version",	"");
	setProperty(prop, "java.class.path", 	"");
	setProperty(prop, "os.name", 			"Plan9");
	setProperty(prop, "os.arch", 			getenv("cputype"));
	setProperty(prop, "os.version", 		"1.0");
	setProperty(prop, "file.separator",		"/");
	setProperty(prop, "path.separator",		" ");
	setProperty(prop, "line.separator",		"\n");
	setProperty(prop, "user.name",			getenv("user"));
	setProperty(prop, "user.home", 			getenv("home"));
	setProperty(prop, "user.dir", 			getwd(buf, sizeof buf));

	return prop;
}

/* *************************************************************
 * 
 * JAVA.LANG.RUNTIME
 *
 */

void
java_lang_Runtime_exitInternal(arg_list args)
{
	Object *	o =	arg_value(args, Object *);
	int 		s = arg_value(args, int);

	USED(o);
	//abort();
	sysExit(s);
}

Object *
java_lang_Runtime_execInternal(arg_list args)
{
	Object *	o 		= arg_value(args, Object *);
	Object *	arg 	= arg_value(args, Object *);
	Object *	env 	= arg_value(args, Object *);

	int arglen = arg ? arg->vars_size / sizeof(int) : -1;
	int envlen = env ? env->vars_size / sizeof(int) : -1;
	char path[1024];
	Object *proc;
	ClassFile *c;

	USED(o, envlen);

	if( arglen < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}

	javaString2CString((Object *)arg->vars[0], path, sizeof(path));

	if( access(path, 0) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}
	
	if( !fork() ){ /* new proc */
		char buf[256]; int i;
		char **argv = (char **)sysCalloc((arglen+1) * sizeof(char *));
		
		/* construction des arguments */
		for(i=0 ; i < arglen ; i++){
			javaString2CString((Object *)arg->vars[i], buf, sizeof buf);
			argv[i] = strdup(buf);
		}		
		/*for(i=0 ; i < envlen ; i++){
			javaString2CString((Object *)env->vars[i], buf, sizeof buf);
			putenv( ???? strdup(buf)); /* il faut faire des choses ici *
		}*/	
		exec(path, argv);
	}

	/* proc père */

	c = lookupClass("java/lang/Process");
	if( !c ){
		throwException(getCurrentFrame(), "java/lang/ClassNotFoundException");
		return 0;
	}
	proc = newObject(c);

	return proc;	
}

Object *
java_lang_Runtime_initializeLinkerInternal(arg_list args)
{
	Object *str = makeJavaString(".");
	USED(args);
	
	return str;
}

Object *
java_lang_Runtime_buildLibName(arg_list args)
{
	Object *str = makeJavaString(".");
	USED(args);
	
	return str;
}

int
java_lang_Runtime_loadFileInternal(arg_list args)
{
	USED(args);
	
	return 1;
}
/* *************************************************************
 * 
 * JAVA.LANG.MATH
 *
 */

double
java_lang_Math_sin(arg_list args)
{
	double v = arg_value(args, double);

	return sin(v);
}

double
java_lang_Math_cos(arg_list args)
{
	double v = arg_value(args, double);

	return cos(v);
}

double
java_lang_Math_tan(arg_list args)
{
	double v = arg_value(args, double);

	return tan(v);
}

double
java_lang_Math_asin(arg_list args)
{
	double v = arg_value(args, double);

	return asin(v);
}

double
java_lang_Math_acos(arg_list args)
{
	double v = arg_value(args, double);

	return acos(v);
}

double
java_lang_Math_atan(arg_list args)
{
	double v = arg_value(args, double);

	return atan(v);
}

double
java_lang_Math_exp(arg_list args)
{
	double v = arg_value(args, double);

	return exp(v);
}

double
java_lang_Math_log(arg_list args)
{
	double v = arg_value(args, double);

	if( v < 0.0 ){
		throwException(getCurrentFrame(), "java/lang/ArithmeticException");
		return 0;
	}

	return log(v);
}

double
java_lang_Math_sqrt(arg_list args)
{
	double v = arg_value(args, double);

	if( v < 0.0 ){
		throwException(getCurrentFrame(), "java/lang/ArithmeticException");
		return 0;
	}

	return sqrt(v);
}

double
java_lang_Math_IEEEremainder(arg_list args)
{
	double v1 = arg_value(args, double);
	double v2 = arg_value(args, double);

	return fmod(v1, v2);
}

double
java_lang_Math_ceil(arg_list args)
{
	double v = arg_value(args, double);

	return ceil(v);
}
	
double
java_lang_Math_floor(arg_list args)
{
	double v = arg_value(args, double);

	return floor(v);
}
	
double
java_lang_Math_rint(arg_list args)
{
	double v = arg_value(args, double);

	return v + 0.5 - fmod(v + 0.5, 1.0); /* du bricolage non !!!! */
}

double
java_lang_Math_atan2(arg_list args)
{
	double v1 = arg_value(args, double);
	double v2 = arg_value(args, double);

	return atan2(v1, v2);
}

double
java_lang_Math_pow(arg_list args)
{
	double v1 = arg_value(args, double);
	double v2 = arg_value(args, double);

	if( (v1 == 0.0 && v2 <= 0.0) || (v1 <= 0.0 && isNaN(v2))){
		throwException(getCurrentFrame(), "java/lang/ArithmeticException");
		return 0.0;
	}

	return pow(v1, v2);
}

/* *************************************************************
 * 
 * JAVA.LANG.DOUBLE
 *
 */

Object *
java_lang_Double_toString(arg_list args)
{
	double d =  arg_value(args, double);

	Object *string;
	char buf[256];
	
	sprint(buf, "%f", d);
	string = makeJavaString(buf);

	return string;
}

Object *
java_lang_Double_valueOf(arg_list args)
{
	Object *string = arg_value(args, Object *);

	Object *db;
	char buf[128];
	double d;

	javaString2CString(string, buf, sizeof buf);
	d = atof(buf);

	/* creation de l'objet Double */
	db = newObject(getClass("java/lang/Double"));

	executeConstructor(db, "(D)V", d);

	return db;
}

vlong
java_lang_Double_doubleToLongBits(arg_list args)
{
	double d = arg_value(args, double);

	union { double d; vlong l; } u;
	
	u.d = d;
	
	return u.l;
}

double
java_lang_Double_longBitsToDouble(arg_list args)
{
	vlong l = arg_value(args, vlong);

	union { double d; vlong l; } u;
	
	u.l = l;
	
	return u.d;
}



/* *************************************************************
 * 
 * JAVA.LANG.FLOAT
 *
 */

Object *
java_lang_Float_toString(arg_list args)
{
	float f =  arg_value(args, float);

	Object *string;
	char buf[256];
	
	sprint(buf, "%f", f);
	string = makeJavaString(buf);

	return string;
}

Object *
java_lang_Float_valueOf(arg_list args)
{
	Object *string = arg_value(args, Object *);

	Object *fl;
	char buf[128];
	float f;

	javaString2CString(string, buf, sizeof buf);
	f = atof(buf);

	/* creation de l'objet Float */
	fl = newObject(getClass("java/lang/Float"));

	executeConstructor(fl, "(F)V", &f);

	return fl;
}

int
java_lang_Float_floatToIntBits(arg_list args)
{
	float f = arg_value(args, float);

	union { float f; int i; } u;
	
	u.f = f;
	
	return u.i;
}

float
java_lang_Float_intBitsToFloat(arg_list args)
{
	int i = arg_value(args, int);

	union { float f; int i; } u;
	
	u.i = i;
	
	return u.f;
}
