/* 
 * parser for *.class file 
*/

#include <u.h>
#include <libc.h>
#include <lock.h>
#include <auth.h>
#include <fcall.h>
#include <bio.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "mzip.h"

extern char *java_class_path;

#define FS_TYPE		1

struct path {
	char *name;
	zip *zip;
};

struct File {
	char *name;
	int type;

	/* read from buffer typically zip files */
	int size;
	uchar *buffer;
	uchar *pt;
	int offset;
	
	/* read from file with bio */
	int fd;
	Biobuf *bio;
};

struct path *class_path;

static ClassFile *loadClassFile(struct File *);
static struct File *openFile(char *);
static void closeFile(struct File *);
static u1 freadU1(struct File *);
static u2 freadU2(struct File *);
static u4 freadU4(struct File *);

static Code_attribute *
ReadCodeAttribute(struct File *, ClassFile *, u1(*)(), u2(*)(), u4(*)());
static Exception_attribute *
ReadExceptionAttribut(struct File *, u1(*)(), u2(*)(), u4(*)());

ClassFile *
loadClass(char *classname)
{
	char buf[512];
	ClassFile *cur_class, *super_class;
	
	cur_class = getClass(classname);
	if( !cur_class ){ /* class not loaded */
		struct path *p = class_path;
		while( p->name ){
			if( !p->zip ){
				sprint(buf,"%s/%s.class", p->name, classname);
				//print("search for <%s>\n", buf);
				cur_class = loadClassFileFromFs(buf);
			}else{
				sprint(buf,"%s.class", classname);
				//print("search for <%s>\n", buf);
				cur_class = loadClassFileFromZip(p->name, p->zip, buf);
			}
			if( cur_class )
				goto loaded;
			p += 1;
		}

	}else
		return cur_class;
	//print("NoClassDefFoundError %s\n", classname);
	return 0;

loaded:
DBG1(chat("[%s loaded]\n", classname);)
	getSuperClassName(buf, cur_class);
	if( !buf[0] ){
		cur_class->r_super = 0;
		goto end;
	}
	super_class = loadClass(buf);
	if( !super_class ){
DBG1(	chat("NoClassDefFoundError for super class %s\n", buf); )
		return 0;
	}
	cur_class->r_super = super_class;
end:
	cur_class->r_state = 0;
	cur_class->r_resolved = 
		(int *) sysCalloc(cur_class->constant_pool_count*sizeof(int));
	cur_class->r_name = (char *) sysMalloc(strlen(classname)+1);
	strcpy(cur_class->r_name, classname);
	addClass(cur_class);
	return cur_class;
}

void makePath(void)
{
	char *path, *path_o, *p, buf[512];
	int count, i, imax;

#define BMAX 5

	path_o = path = strdup(java_class_path);

	 /* eat spaces at the begining */
	while( *path && (*path == ' ' || *path == '\t') ) path++;

	/* eat spaces at the end */
	p = path + strlen(path)-1;
	while( p > path && (*p == ' ' || *p == '\t') ) p--; 
	*(p+1) = 0;

	class_path = (struct path*)sysCalloc(BMAX * sizeof(struct path));
	if( !class_path )
		panic("calloc failed class_path in makePath()\n");

	count = getNbPath(path);
	for(i = 0, imax = BMAX ; i < count ; i++){
		getPath(buf, path, i);

		if( i >= imax ){
			struct path *p;
			p = (struct path *)sysCalloc((imax + BMAX) * sizeof(struct path));
			if( !p )
				panic("can't grow class_path makePath()\n");
			memcpy((char *)p, (char *)class_path, imax * sizeof(struct path));
			free(class_path);
			class_path = p;
		}
			
		if( strstr(buf, ".zip") ){
			zip *z = zipOpen(buf);
			if( !z )
				panic("problems with %s zip file\n", buf);
			class_path[i].name = strdup(buf);
			class_path[i].zip = z;
		}else{
			class_path[i].name = strdup(buf);
			class_path[i].zip = 0;
		}
	}
	free(path_o);

}


ClassFile *
loadClassFileFromZip(char *zip_file, zip *z, char *filename)
{
	ClassFile *class;
	struct File *f;
    zipdir *dp;

	dp = zipLookup(z, filename);
	if( !dp )
		return 0;

	f = openFile(zip_file);
	if( !f )
		panic("openFile failed\n");
	
	{
    	char locbuf[LOCHDRSIZE];
    	long off;

   	 	/* Seek to beginning of LOC header */
    	if( Bseek(f->bio, dp->off, 0) < 0 )
			panic("Bseek 1 loadClassFileFromZip\n");

		/* Read LOC header */
    	if( Bread(f->bio, locbuf, LOCHDRSIZE) != LOCHDRSIZE )
			panic("Bread 2 in loadClassFileFromZip\n");
			
	    /* Verify signature */
	    if( strncmp(locbuf, LOCSIG, SIGSIZE) != 0 )
			panic("Invalid LOC header signature loadClassFileFromZip\n");

	    /* Get offset of file data */
 		off = dp->off + LOCHDRSIZE + LOCNAM(locbuf) + LOCEXT(locbuf);
	    if (off + dp->len > z->cenoff) 
			panic("Invalid LOC header in loadClassFileFromZip\n");
	
  		/* Seek to file data */
	    if( Bseek(f->bio, off, 0) < 0 )
			panic("Bseek offset in loadClassFileFromZip\n");
	}

	f->type = FS_TYPE;
	class = loadClassFile(f);
	closeFile(f);

	return class;
}

ClassFile *
loadClassFileFromFs(char *filename)
{	
	struct File *f;
	ClassFile *class;

	f = openFile(filename);
	if( !f )
		return 0;

	f->type = FS_TYPE;
	class = loadClassFile(f);
	closeFile(f);

	return class;
}

static ClassFile *
loadClassFile(struct File *f)
{
	int i, j;
	char buf[1024];
	cp_info *cp;
	ClassFile *class;
	u1 (*preadU1)(struct File *);
	u2 (*preadU2)(struct File *);
	u4 (*preadU4)(struct File *);

	SET(preadU1, preadU2, preadU4);

	switch(f->type){
		case FS_TYPE:
			preadU1 = freadU1;
			preadU2 = freadU2;
			preadU4 = freadU4;
			break;
		default:
			panic("bad type in loadClassFile()\n");
	}

	class = (ClassFile *) sysMalloc(sizeof(ClassFile));
	if ( !class )
		panic("Can't malloc a new ClassFile...");

	class->magic 			= (*preadU4)(f);
	class->minor_version 	= (*preadU2)(f); 
	class->major_version 	= (*preadU2)(f);

	if ( class->magic != CLASSEFILE_MAGIC ) 
		panic("Bad magic number %d", class->magic);

	if ( class->major_version != 45 || class->minor_version != 3 )
		panic("Version Number not suported.");


	/*
	 * Read Constant Pool area
	*/

	class->constant_pool_count = (*preadU2)(f);
  	
	if ( class->constant_pool_count ) {
		class->constant_pool = 
			(cp_info *)sysMalloc(class->constant_pool_count*sizeof(cp_info));
		if (!class->constant_pool)
			panic("malloc failed for constant_pool");
	}
	for (i=1 ; i < class->constant_pool_count ; i++) {
		cp = &class->constant_pool[i];
		cp->tag = (*preadU1)(f);
	
		switch (cp->tag) {
    
			case CONSTANT_Asciz:
				cp->u.a.length = (*preadU2)(f);
				cp->u.a.bytes = (u1 *)sysMalloc(cp->u.a.length*sizeof(u1));
				for (j=0 ; j < cp->u.a.length ; j++)
					cp->u.a.bytes[j] = (*preadU1)(f);
				break;
      
			case CONSTANT_Class:
				cp->u.c.name_index = (*preadU2)(f);
				break;
			case CONSTANT_String:
				cp->u.s.string_index = (*preadU2)(f);
				break;
      
			case CONSTANT_Fieldref:
			case CONSTANT_Methodref:
			case CONSTANT_InterfaceMethodref:
				cp->u.m.class_index 		= (*preadU2)(f);
				cp->u.m.name_and_type_index = (*preadU2)(f);
				break;

			case CONSTANT_NameAndType:
				cp->u.n.name_index 		= (*preadU2)(f);
				cp->u.n.signature_index = (*preadU2)(f);
				break;

			case CONSTANT_Float:
			case CONSTANT_Integer:
				cp->u.i.bytes = (*preadU4)(f);
				break;	
      
			case CONSTANT_Long:
			case CONSTANT_Double:
				i++;
#ifdef LITTLE_ENDIAN
					cp->u.l.hight_bytes = (*preadU4)(f);
					cp->u.l.low_bytes 	= (*preadU4)(f);
#else
					cp->u.l.low_bytes 	= (*preadU4)(f);
					cp->u.l.hight_bytes = (*preadU4)(f);
#endif					
				break;

			default: 
				panic("Illegal constant pool type %d\n", cp->tag); 
		}
	} /* end of constant pool read loop*/

	class->access_flags 	= (*preadU2)(f);
	class->this_class 		= (*preadU2)(f);
	class->super_class 		= (*preadU2)(f);

	/*
	 * Read interfaces area
	 */

	class->interfaces_count = (*preadU2)(f);
 	if ( class->interfaces_count ) {
		class->interfaces = (u2 *) sysMalloc(class->interfaces_count * sizeof(u2));
		if( !class->interfaces )
			panic("malloc failed for interface");
	}
	
	for(i=0 ; i < class->interfaces_count ; i++)
		class->interfaces[i] = (*preadU2)(f);

	/*
	 * Read Fields area
	 */

	class->fields_count = (*preadU2)(f);
	if ( class->fields_count ) {
		class->fields = 
			(field_info *) sysMalloc(class->fields_count*sizeof(field_info));
		if ( !class->fields )
			panic("malloc failed for fields");
	}
	
	for (i=0 ; i < class->fields_count ; i++) {
		u2 attributes_count;
		class->fields[i].access_flag 		= (*preadU2)(f);
		class->fields[i].name_index 		= (*preadU2)(f);
		class->fields[i].signature_index 	= (*preadU2)(f);

		attributes_count = class->fields[i].attributes_count = (*preadU2)(f);
		if( attributes_count ) {
			class->fields[i].attributes = 
				(attribute_info *)sysMalloc(attributes_count*sizeof(attribute_info));
    			if ( !class->fields[i].attributes ) 
				panic("malloc failed for field attribute");
		}
		for(j=0 ; j < attributes_count ; j++){
			int length, k;
			class->fields[i].attributes[j].attribute_name = (*preadU2)(f);
			length = class->fields[i].attributes[j].attribute_length = (*preadU4)(f);
			if( length ){
				class->fields[i].attributes[j].info = (u1 *) sysMalloc(length*sizeof(u1));
				if ( !class->fields[i].attributes[j].info )
					panic("malloc failed for fields attributes info");
			}
			for(k=0 ; k < length ; k++)
				class->fields[i].attributes[j].info[k] = (*preadU1)(f);
		}
	}

	/*
	 * Read Methods area
	*/

	class->methods_count = (*preadU2)(f);
	if ( class->methods_count ) {
		class->methods = 
			(method_info *) sysMalloc(class->methods_count*sizeof(method_info));
  		if ( !class->methods )
			panic("malloc failed for methods");
	}
	for (i=0 ; i < class->methods_count ; i++) {
		u2 count;
		class->methods[i].access_flag = (*preadU2)(f);
		/* test ACC_PRIVATE or ACC_PUBLIC */
		
		class->methods[i].name_index = (*preadU2)(f);
		GetCpString(buf, class->methods[i].name_index, class);
		
		class->methods[i].signature_index = (*preadU2)(f);
		
		count = class->methods[i].attributes_count = (*preadU2)(f);
    	class->methods[i].attributes = 
			(struct method_attribute *) sysMalloc(count*sizeof(struct method_attribute));

		for(j=0 ; j < count ; j++){
			u4 length;
			u2 attribute_name;

			attribute_name = (*preadU2)(f);
			GetCpString(buf, attribute_name, class);
			
			length = (*preadU4)(f);

			if ( !strcmp(buf, "Code") ){
				class->methods[i].attributes[j].type = ATTR_CODE;
				class->methods[i].attributes[j].u.ca = ReadCodeAttribute(f, class, 
									preadU1, preadU2, preadU4);
				class->methods[i].attributes[j].u.ca->attribute_name_index = attribute_name;
				class->methods[i].attributes[j].u.ca->attribute_length = length;
			} else if ( !strcmp(buf, "Exceptions") ){
				class->methods[i].attributes[j].type = ATTR_EXCEPTION;
				class->methods[i].attributes[j].u.ea = ReadExceptionAttribut(f, 
															preadU1, preadU2, preadU4);
			}else
				panic("not Code or Exception attribut");
		}
	}
	
	/*
	 * Read Attribute area
	*/

	class->attributes_count = (*preadU2)(f);
	if ( class->attributes_count ) {
		class->attributes = 
			(attribute_info *) sysMalloc(class->attributes_count*sizeof(attribute_info));
  		if ( !class->attributes_count )
			panic("malloc failed for attributes");
	}
	for(i=0; i < class->attributes_count; i++){
		int length, file_index;
		class->attributes[i].attribute_name = (*preadU2)(f);
		GetCpString(buf, class->attributes[i].attribute_name, class);
		length = class->attributes[i].attribute_length = (*preadU4)(f);
		if ( strcmp(buf, "SourceFile") || length != 2 ){
			//print("bad attribute, must be \"SourceFile\" but %s\n", buf);
			class->attributes[i].info = (u1 *) sysMalloc(length*sizeof(u1));
			for(j=0 ; j < length ; j++)
				class->attributes[i].info[j] = (*preadU1)(f);
		}else{
			file_index = (*preadU2)(f);
			GetCpString(buf, file_index, class);
			class->r_file = strdup(buf);
		}
	}

	/*if( (f->type == ZIP_TYPE) && (f->offset != f->size) )
		panic("Il reste des truc non lus loadClassFile()\n");*/
	/*if( (f->type == ZIP_TYPE) )
		print("read = %d bytes\n", f->pt-f->buffer);*/

	return class;
}



static Code_attribute *
ReadCodeAttribute(struct File *f, ClassFile *class, 
						u1 (*readU1)(struct File *),
						u2 (*readU2)(struct File *), 
						u4 (*readU4)(struct File *))
{
	Code_attribute *ca;
	char buf[1024];
	int i, j;

	ca = (Code_attribute *) sysMalloc(sizeof(Code_attribute));
	if (!ca) 
		panic("malloc failed for Code_attribut");

	ca->max_stack = (*readU2)(f);
	ca->max_locals = (*readU2)(f); 
	ca->code_length = (*readU4)(f);

	ca->code = (u1 *) sysMalloc(ca->code_length * sizeof(u1));
	for (i=0; i < ca->code_length; i++)
		ca->code[i] = (*readU1)(f);
	
	ca->exception_table_length = (*readU2)(f); 
	ca->exception_table = 
		(exception *) sysMalloc(ca->exception_table_length*sizeof(exception));
	for (i=0; i < ca->exception_table_length; i++) {
		ca->exception_table[i].start_pc = (*readU2)(f);
		ca->exception_table[i].end_pc = (*readU2)(f);
		ca->exception_table[i].handler_pc = (*readU2)(f);
		ca->exception_table[i].catch_type = (*readU2)(f);
	}
	ca->attribute_count = (*readU2)(f);
	ca->attribute = 
		(struct code_attribute *)sysMalloc(ca->attribute_count*sizeof(struct code_attribute));
	for (i=0; i < ca->attribute_count ; i++) {
		/*ca->attribute[i].attribute_name = (*readU2)(f);
		GetCpString(buf, ca->attribute[i].attribute_name, class);
		ca->attribute[i].attribute_length= (*readU4)(f); */		
		int attribute_name =  (*readU2)(f), length = (*readU4)(f);

		GetCpString(buf, attribute_name, class);

		if ( !strcmp(buf, "LineNumberTable" ) ){
			ca->attribute[i].type = ATTR_LINE_NUMBER;
			ca->attribute[i].u.la.attribute_name_index = attribute_name;
			ca->attribute[i].u.la.attribute_length = length;
			ca->attribute[i].u.la.line_number_table_length = (*readU2)(f);
			ca->attribute[i].u.la.line_number_table = 
				(struct line_number *) sysMalloc(
					ca->attribute[i].u.la.line_number_table_length *
					sizeof(struct line_number));
			for(j=0 ; j < ca->attribute[i].u.la.line_number_table_length ; j++){
				ca->attribute[i].u.la.line_number_table[j].start_pc = (*readU2)(f);
				ca->attribute[i].u.la.line_number_table[j].line_number = (*readU2)(f);
			}	

		}else if ( !strcmp(buf, "LocalVariableTable" ) ){
			ca->attribute[i].type = ATTR_LOCAL_VARIABLE;
			ca->attribute[i].u.va.attribute_name_index = attribute_name;
			ca->attribute[i].u.va.attribute_length = length;
			ca->attribute[i].u.va.local_variable_table_length = (*readU2)(f);
			ca->attribute[i].u.va.local_variable_table = 
				(struct local_variable *) sysMalloc(
					ca->attribute[i].u.va.local_variable_table_length *
					sizeof(struct local_variable));
			for(j=0 ; j < ca->attribute[i].u.va.local_variable_table_length ; j++){
				ca->attribute[i].u.va.local_variable_table[j].start_pc = (*readU2)(f);
				ca->attribute[i].u.va.local_variable_table[j].length = (*readU2)(f);
				ca->attribute[i].u.va.local_variable_table[j].name_index = (*readU2)(f);
				ca->attribute[i].u.va.local_variable_table[j].signature_index = (*readU2)(f);
				ca->attribute[i].u.va.local_variable_table[j].slot = (*readU2)(f);
			}	

		}else
			panic("not LocalVariableTable or LineNumberTable attribute");

		/*ca->attribute[i].info = (u1 *) sysMalloc(ca->attribute[i].attribute_length);
		for (j=0; j < ca->attribute[i].attribute_length; j++)
			ca->attribute[i].info[j] = (*readU1)(f);*/
	}
	return ca;
}

static Exception_attribute *
ReadExceptionAttribut(struct File *f, u1 (*readU1)(struct File *),
									  u2 (*readU2)(struct File *), 
									  u4 (*readU4)(struct File *))
{
	Exception_attribute *ea;
	int count, i;
	USED(readU1, readU4);

	ea = (Exception_attribute *) sysMalloc(sizeof(Exception_attribute));
	if (!ea) 
		panic("malloc failed for Exception_attribute");

	count = ea->number_of_exception = (*readU2)(f);
	if ( count ) {
		ea->exception_index_table = (u2 *) sysMalloc(count*sizeof(u2));
		if (!ea->exception_index_table)
			panic("malloc failed for exception table");
		for (i=0; i < count; i++) 
			ea->exception_index_table[i] = (*readU2)(f);
	}
	return ea;
}


/*
 * Functions to read class file
 */


static struct File *
openFile(char *filename)
{
	struct File *f;
	Biobuf * bio;

	bio = Bopen(filename, OREAD);
	if ( !bio ){
		return 0;
	}

	f = (struct File *) sysMalloc(sizeof(struct File));
	if ( !f )
		panic("malloc failed for f in openFile");
	f->name = filename;
	f->bio = bio;

	return f;
}

static void
closeFile(struct File *f)
{
	Bterm(f->bio);
	free(f);
}

static u1
freadU1(struct File *f)
{
	u1 a;

	if( Bread(f->bio, &a, sizeof(u1)) != sizeof(u1) )
		panic("readU1\n");
	return a;
}

static u2
freadU2(struct File *f)
{
	u2 a;

	if( Bread(f->bio, &a, sizeof(u2)) != sizeof(u2) )
		panic("readU2\n");

#ifdef LITTLE_ENDIAN
		convU2(&a);
#endif

	return a;
}

static u4
freadU4(struct File *f)
{
	u4 a;

	if( Bread(f->bio, &a, sizeof(u4)) != sizeof(u4) )
		panic("readU4\n");

#ifdef LITTLE_ENDIAN
		convU4(&a);
#endif
	return a;
}


/*
static struct File *
openFile( char *filename )
{
	struct File *f;
	char ap[DIRLEN]; Dir edir;
	int fd;

	fd = open(filename, OREAD);
	if ( fd < 0 ) 
		return 0;

	f = (struct File *) sysMalloc(sizeof(struct File));
	if ( !f )
		panic("malloc failed for f in openFile");
	f->name = filename;

	fstat(fd, ap);
	convM2D(ap, &edir);
	f->size = edir.length;
	
	f->buffer = (char *) sysMalloc(f->size);
	if ( !f->buffer )
		panic("malloc failed for buffer in openFile()");
	f->pt = f->buffer;

	if( read(fd, f->buffer, f->size) < f->size )
		panic("can't read in %s", f->name);
	
	close(fd);

	return f;
}*/


/*static void
closeFile( struct File *f )
{
	free(f->buffer);
	free(f);
}*/
