#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "thread.h"

extern int java_max_stack;

StackFrame *current_stack;
Frame *current_frame;

#define DEFAULT_STACK_SIZE	128

static void pushFrame(Frame *f)
{
	int nsize;

	nsize = current_stack->size + sizeof(Frame) +
			(FRAME_MAX_STACK(f) + FRAME_MAX_LOCALS(f)) * sizeof(int);
	if( nsize > java_max_stack )
		panic("stack overflow size = %d bytes\n", nsize);

	if( current_stack->pt + 1 >= current_stack->max ){
		current_stack->stack = 
			(Frame **)realloc(current_stack->stack,
								(current_stack->max + DEFAULT_STACK_SIZE) 
								* sizeof(Frame *));
		if( !current_stack->stack )
			panic("can't grow stack in pushFrame()\n");
		current_stack->max += DEFAULT_STACK_SIZE;
	}
		
	current_stack->stack[++current_stack->pt] = f;
	current_stack->size = nsize;
}
	
static Frame *popFrame(void)
{
	if( current_stack->pt <= -1 )
		panic("stack frame underflow in popFrame()\n");
	current_stack->size -= sizeof(Frame) +
			(FRAME_MAX_STACK(current_frame) + FRAME_MAX_LOCALS(current_frame)) *
			sizeof(int);
	return current_stack->stack[current_stack->pt--];
}

Frame *
getCurrentFrame(void)
{
	return current_stack->stack[current_stack->pt];
}


StackFrame *
initGlobalStack(void)
{
	StackFrame *sf;

	sf = (StackFrame *)sysCalloc(sizeof(StackFrame));
	if( !sf )
		panic("can't allocate new stack frame in initGlobalStack()\n");
	sf->stack = (Frame **)sysCalloc(DEFAULT_STACK_SIZE * sizeof(Frame *));
	sf->max = DEFAULT_STACK_SIZE;
	sf->pt = -1;
	sf->size = 0;

DBG1(chat("[global stack initialized]\n");)

	return sf;
}
void
freeFrame(Frame *frame)
{
	Frame *f = popFrame();
	if( frame != f )
		panic("stack frame problem...\n");
	free(f->opstack);
	free(f->locals);
	free(f);
	current_frame = current_stack->stack[current_stack->pt];
}

Frame *
newFrame(MethodBlock *mb, Frame *invoker)
{
	Frame *frame;

 	frame = (Frame *)sysMalloc(sizeof(Frame));
	if( !frame )
		panic("can't malloc new frame in newFrame()\n");

	frame->class 		= mb->class;
	frame->mb 			= mb;
	frame->name 		= mb->name;
	frame->pc 			= MB_CODE(mb);
	frame->opstack 		= (int *) sysMalloc(MB_MAX_STACK(mb) * sizeof(int));
	frame->locals 		= (int *) sysMalloc(MB_MAX_LOCALS(mb) * sizeof(int));
	frame->opstack_pt 	= -1;
	frame->invoker 		= invoker;
	frame->exception 	= 0;
	frame->obj_lock 	= 0;
	
	pushFrame(frame);
	current_frame = current_stack->stack[current_stack->pt];

	return frame;
}

Frame *
createFrame(StackFrame *sf, MethodBlock *mb, Frame *invoker)
{
	Frame *frame;
	StackFrame *tmp_stack;

 	frame = (Frame *)sysMalloc(sizeof(Frame));
	if( !frame )
		panic("can't malloc new frame in newFrame()\n");

	frame->class 		= mb->class;
	frame->mb 			= mb;
	frame->name 		= mb->name;
	frame->pc 			= MB_CODE(mb);
	frame->opstack 		= (int *) sysMalloc(MB_MAX_STACK(mb) * sizeof(int));
	frame->locals 		= (int *) sysMalloc(MB_MAX_LOCALS(mb) * sizeof(int));
	frame->opstack_pt 	= -1;
	frame->invoker 		= invoker;
	frame->exception 	= 0;
	frame->obj_lock 	= 0;
	
	tmp_stack = current_stack;
	current_stack = sf;
	pushFrame(frame);
	current_stack = tmp_stack;

	return frame;
}
Frame *
newFrameFromName(ClassFile *class, char *method_name, char *sign)
{
	int m; char buf[256];
	Frame *frame;
	MethodBlock *mb;

	sprint(buf, "%s%s", method_name, sign);

	if( !strcmp("<clinit>()V", buf) ){
		for (m=class->r_n_methods-1; m >= 0 ; m--){
			mb = &class->r_methods[m];
			if( mb->class != class )
				break;
			if( !strcmp(buf, mb->name) )
				goto found;
		}	
	}else{
		for (m=class->r_n_methods-1; m >= 0 ; m--){
			mb = &class->r_methods[m];
			if( !strcmp(buf, mb->name) )
				goto found;
		}
	}
	return 0;

found:
 	frame = (Frame *)sysMalloc(sizeof(Frame));
	if( !frame )
		panic("can't malloc new frame in newFrame()\n");

	frame->class 		= class;
	frame->mb 			= mb;
	frame->name 		= mb->name;
	frame->pc 			= MB_CODE(mb);
	frame->opstack 		= (int *) sysMalloc(MB_MAX_STACK(mb) * sizeof(int));
	frame->locals 		= (int *) sysMalloc(MB_MAX_LOCALS(mb) * sizeof(int));
	frame->opstack_pt 	= -1;
	frame->invoker 		= 0;
	frame->exception 	= 0;
	frame->obj_lock 	= 0;

	pushFrame(frame);
	current_frame = current_stack->stack[current_stack->pt];

	return frame;
}

Frame *
createFrameFromName(StackFrame *sf, ClassFile *class, char *method_name, char *sign)
{
	int m; char buf[256];
	Frame *frame;
	MethodBlock *mb;
	StackFrame *tmp_stack;

	sprint(buf, "%s%s", method_name, sign);

	if( !strcmp("<clinit>()V", buf) ){
		for (m=class->r_n_methods-1; m >= 0 ; m--){
			mb = &class->r_methods[m];
			if( mb->class != class )
				break;
			if( !strcmp(buf, mb->name) )
				goto found;
		}	
	}else{
		for (m=class->r_n_methods-1; m >= 0 ; m--){
			mb = &class->r_methods[m];
			if( !strcmp(buf, mb->name) )
				goto found;
		}
	}
	return 0;

found:
 	frame = (Frame *)sysMalloc(sizeof(Frame));
	if( !frame )
		panic("can't malloc new frame in newFrame()\n");

	frame->class 		= class;
	frame->mb 			= mb;
	frame->name 		= mb->name;
	frame->pc 			= MB_CODE(mb);
	frame->opstack 		= (int *) sysMalloc(MB_MAX_STACK(mb) * sizeof(int));
	frame->locals 		= (int *) sysMalloc(MB_MAX_LOCALS(mb) * sizeof(int));
	frame->opstack_pt 	= -1;
	frame->invoker 		= 0;
	frame->exception 	= 0;
	frame->obj_lock 	= 0;

	tmp_stack = current_stack;
	current_stack = sf;
	pushFrame(frame);
	current_stack = tmp_stack;

	return frame;
}

Frame *findInvokerFrame(Frame *f)
{
	if( f->invoker )
		return f->invoker;
	
	if( current_stack->pt )
		return current_stack->stack[current_stack->pt - 1];

	return 0;
}


#ifdef DBG_1
void dumpStackFrame()
{
	Frame *f = getCurrentFrame();

	for(;;){
		dumpFrame(f);
		if( !strcmp(f->name, "main([Ljava/lang/String;)V") )
			break;
		f = f->invoker;
	}

}
#endif

void fillInStackTrace(Object *thr)
{	
	ClassFile *class_thr;
	Object *tab;
	int count, i;

	/* sauter les frame d'init de l'objet exception */
	
	class_thr = getClass("java/lang/Throwable");
	for(i=current_stack->pt ; i >= 0 ; i--)
		if( !isSubClassOf(current_stack->stack[i]->class, class_thr) )
			break;
	/* compter les frames utiles */
	
	count = i + 1;

	/* allocation du tableau contenant MethodBlock + pc */
	
	tab = newArray(T_INT, count * 2);
	
	/* sauver dans le tableau la paire MethodBlock + pc */
	
	for(count = 0; i >= 0 ; i--){
		Frame *f = current_stack->stack[i];
		tab->vars[count++] = (int)f->mb;
		tab->vars[count++] = pc2LineCode(f->mb, f->pc - FRAME_CA(f)->code);
	}
	thr->vars[0] = (int)tab;
}

int
pc2LineCode(MethodBlock *mb, int pc_offset)
{
	Code_attribute *ca = mb->ca;
	LineNumberTable_attribute *la;
	int i, j;

/*	print("pc2LineNumber():\nTable for %s.%s\n", mb->class->r_name, mb->name);
	for(i=0 ; i < ca->attribute_count ; i++){
		if( ca->attribute[i].type == ATTR_LINE_NUMBER ){
			la = &ca->attribute[i].u.la;
			for(j=0 ; j < la->line_number_table_length ; j++){
				print("pc = %d :: start_pc = %d -> %d\n", 
					pc_offset, 
					la->line_number_table[j].start_pc, 
					la->line_number_table[j].line_number);
			}
		}
	}*/

	for(i=0 ; i < ca->attribute_count ; i++){
		if( ca->attribute[i].type == ATTR_LINE_NUMBER ){
			la = &ca->attribute[i].u.la;
			if( la->line_number_table_length == 1 )
				return la->line_number_table[0].line_number;
			for(j=1 ; j < la->line_number_table_length ; j++){
				if( pc_offset >= la->line_number_table[j-1].start_pc && 
					pc_offset <= la->line_number_table[j].start_pc ){
					return la->line_number_table[j-1].line_number;
				}
			}
			return la->line_number_table[j-1].line_number;
		}
	}
	
	return -1;
}
