/*
 * native.c 
 */
#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "native.h"

struct function {
	char *name;
	union t {
		void 		(*v)(arg_list);
		int 		(*i)(arg_list);
		vlong		(*l)(arg_list);
		float 		(*f)(arg_list);
		double 		(*d)(arg_list);
		Object *	(*po)(arg_list);
	}u;
};

typedef void (*fvoid)();

struct function Functions[] = {

	/*
	 * JAVA.LANG.*
	 */

	{ "java/lang/Thread.currentThread", {(fvoid)java_lang_Thread_currentThread}},
	{ "java/lang/Thread.setPriority0", 	{(fvoid)java_lang_Thread_setPriority0}},
	{ "java/lang/Thread.start", 		{(fvoid)java_lang_Thread_start}},
	{ "java/lang/Thread.yield", 		{(fvoid)java_lang_Thread_yield}},
	{ "java/lang/Thread.sleep", 		{(fvoid)java_lang_Thread_sleep}},
	{ "java/lang/Thread.stop0", 		{(fvoid)java_lang_Thread_stop0}},
	{ "java/lang/Thread.suspend0", 		{(fvoid)java_lang_Thread_suspend0}},
	{ "java/lang/Thread.resume0", 		{(fvoid)java_lang_Thread_resume0}},
	{ "java/lang/Thread.isAlive", 		{(fvoid)java_lang_Thread_isAlive}},

	{ "java/lang/System.currentTimeMillis", {(fvoid)java_lang_System_currentTimeMillis}},
	{ "java/lang/System.arraycopy", 		{(fvoid)java_lang_System_arraycopy}},
	{ "java/lang/System.initProperties", 	{(fvoid)java_lang_System_initProperties}},

	{ "java/lang/Math.sin", 			{(fvoid)java_lang_Math_sin}},
	{ "java/lang/Math.cos", 			{(fvoid)java_lang_Math_cos}},
	{ "java/lang/Math.tan", 			{(fvoid)java_lang_Math_tan}},
	{ "java/lang/Math.asin", 			{(fvoid)java_lang_Math_asin}},
	{ "java/lang/Math.acos", 			{(fvoid)java_lang_Math_acos}},
	{ "java/lang/Math.atan", 			{(fvoid)java_lang_Math_atan}},
	{ "java/lang/Math.exp", 			{(fvoid)java_lang_Math_exp}},
	{ "java/lang/Math.log", 			{(fvoid)java_lang_Math_log}},
	{ "java/lang/Math.sqrt", 			{(fvoid)java_lang_Math_sqrt}},
	{ "java/lang/Math.IEEEremainder", 	{(fvoid)java_lang_Math_IEEEremainder}},
	{ "java/lang/Math.ceil",			{(fvoid)java_lang_Math_ceil}},
	{ "java/lang/Math.floor", 			{(fvoid)java_lang_Math_floor}},
	{ "java/lang/Math.rint", 			{(fvoid)java_lang_Math_rint}},
	{ "java/lang/Math.atan2", 			{(fvoid)java_lang_Math_atan2}},
	{ "java/lang/Math.pow", 			{(fvoid)java_lang_Math_pow}},

	{ "java/lang/Object.wait", 		{(fvoid)java_lang_Object_wait}},
	{ "java/lang/Object.notify", 	{(fvoid)java_lang_Object_notify}},
	{ "java/lang/Object.hashCode", 	{(fvoid)java_lang_Object_hashCode}},
	{ "java/lang/Object.clone", 	{(fvoid)java_lang_Object_clone}},
	{ "java/lang/Object.getClass", 	{(fvoid)java_lang_Object_getClass}},
	{ "java/lang/Object.notifyAll", {(fvoid)java_lang_Object_notifyAll}},

	{ "java/lang/Double.toString", 			{(fvoid)java_lang_Double_toString}},
	{ "java/lang/Double.valueOf", 			{(fvoid)java_lang_Double_valueOf}},
	{ "java/lang/Double.doubleToLongBits", 	{(fvoid)java_lang_Double_doubleToLongBits}},
	{ "java/lang/Double.longBitsToDouble", 	{(fvoid)java_lang_Double_longBitsToDouble}},

	{ "java/lang/Float.toString", 		{(fvoid)java_lang_Float_toString}},
	{ "java/lang/Float.valueOf", 		{(fvoid)java_lang_Float_valueOf}},
	{ "java/lang/Float.floatToIntBits", {(fvoid)java_lang_Float_floatToIntBits}},
	{ "java/lang/Float.intBitsToFloat", {(fvoid)java_lang_Float_intBitsToFloat}},


	{ "java/lang/Throwable.fillInStackTrace", {(fvoid)java_lang_Throwable_fillInStackTrace}},
	{ "java/lang/Throwable.printStackTrace0", {(fvoid)java_lang_Throwable_printStackTrace0}},

	{ "java/lang/Class.forName", 		{(fvoid)java_lang_Class_forName}},
	{ "java/lang/Class.newInstance", 	{(fvoid)java_lang_Class_newInstance}},
	{ "java/lang/Class.getSuperclass", 	{(fvoid)java_lang_Class_getSuperclass}},
	{ "java/lang/Class.getName", 		{(fvoid)java_lang_Class_getName}},
	{ "java/lang/Class.getInterfaces", 	{(fvoid)java_lang_Class_getInterfaces}},
	{ "java/lang/Class.isInterface", 	{(fvoid)java_lang_Class_isInterface}},
	{ "java/lang/Class.getClassLoader", {(fvoid)java_lang_Class_getClassLoader}},

	{ "java/lang/Runtime.exitInternal", {(fvoid)java_lang_Runtime_exitInternal}},
	{ "java/lang/Runtime.execInternal", {(fvoid)java_lang_Runtime_execInternal}},
	{ "java/lang/Runtime.initializeLinkerInternal", {(fvoid)java_lang_Runtime_initializeLinkerInternal}},
	{ "java/lang/Runtime.buildLibName", {(fvoid)java_lang_Runtime_buildLibName}},
	{ "java/lang/Runtime.loadFileInternal", {(fvoid)java_lang_Runtime_loadFileInternal}},

	/*
	 * JAVA.IO.*
	 */

	{ "java/io/FileDescriptor.initSystemFD", {(fvoid)java_io_FileDescriptor_initSystemFD}},
	{ "java/io/FileDescriptor.valid", {(fvoid)java_io_FileDescriptor_valid}},

	{ "java/io/RandomAccessFile.open", 			{(fvoid)java_io_RandomAccessFile_open}},
	{ "java/io/RandomAccessFile.read", 			{(fvoid)java_io_RandomAccessFile_read}},
	{ "java/io/RandomAccessFile.readBytes", 	{(fvoid)java_io_RandomAccessFile_readBytes}},
	{ "java/io/RandomAccessFile.write", 		{(fvoid)java_io_RandomAccessFile_write}},
	{ "java/io/RandomAccessFile.writeBytes", 	{(fvoid)java_io_RandomAccessFile_writeBytes}},
	{ "java/io/RandomAccessFile.getFilePointer", {(fvoid)java_io_RandomAccessFile_getFilePointer}},
	{ "java/io/RandomAccessFile.seek", 			{(fvoid)java_io_RandomAccessFile_seek}},
	{ "java/io/RandomAccessFile.length", 		{(fvoid)java_io_RandomAccessFile_length}},
	{ "java/io/RandomAccessFile.close", 		{(fvoid)java_io_RandomAccessFile_close}},

	{ "java/io/File.exists0", 		{(fvoid)java_io_File_exists0}},
	{ "java/io/File.canRead0", 		{(fvoid)java_io_File_canRead0}},
	{ "java/io/File.canWrite0", 	{(fvoid)java_io_File_canWrite0}},
	{ "java/io/File.isFile0", 		{(fvoid)java_io_File_isFile0}},
	{ "java/io/File.isDirectory0", 	{(fvoid)java_io_File_isDirectory0}},
	{ "java/io/File.length0", 		{(fvoid)java_io_File_length0}},
	{ "java/io/File.mkdir0", 		{(fvoid)java_io_File_mkdir0}},
	{ "java/io/File.delete0", 		{(fvoid)java_io_File_delete0}},
	{ "java/io/File.renameTo0", 	{(fvoid)java_io_File_renameTo0}},
	{ "java/io/File.isAbsolute", 	{(fvoid)java_io_File_isAbsolute}},
	{ "java/io/File.list0", 		{(fvoid)java_io_File_list0}},
	{ "java/io/File.lastModified0", {(fvoid)java_io_File_lastModified0}},

	{ "java/io/FileInputStream.open", 		{(fvoid)java_io_FileInputStream_open}},
	{ "java/io/FileInputStream.read", 		{(fvoid)java_io_FileInputStream_read}},
	{ "java/io/FileInputStream.readBytes", 	{(fvoid)java_io_FileInputStream_readBytes}},
	{ "java/io/FileInputStream.skip", 		{(fvoid)java_io_FileInputStream_skip}},
	{ "java/io/FileInputStream.available",	{(fvoid)java_io_FileInputStream_available}},
	{ "java/io/FileInputStream.close", 		{(fvoid)java_io_FileInputStream_close}},

	{ "java/io/FileOutputStream.open", 		{(fvoid)java_io_FileOutputStream_open}},
	{ "java/io/FileOutputStream.write", 	{(fvoid)java_io_FileOutputStream_write}},
	{ "java/io/FileOutputStream.writeBytes",{(fvoid)java_io_FileOutputStream_writeBytes}},
	{ "java/io/FileOutputStream.close",		{(fvoid)java_io_FileOutputStream_close}},

	/*
	 * JAVA.UTIL.*
	 */

	{ "java/util/Date.expand",			{(fvoid)java_util_Date_expand}},
	{ "java/util/Date.computeValue",	{(fvoid)java_util_Date_computeValue}},
	{ "java/util/Date.toString",		{(fvoid)java_util_Date_toString}},
	{ "java/util/Date.toLocalString",	{(fvoid)java_util_Date_toLocalString}},
	{ "java/util/Date.toGMTString",		{(fvoid)java_util_Date_toGMTString}},

	/*
	 * JAVA.NET.*
	 */

	{ "java/net/PlainSocketImpl.socketCreate",	{(fvoid)java_net_PlainSocketImpl_socketCreate}},

	{ "java/net/InetAddress.makeAnyLocalAddress",{(fvoid)java_net_InetAdress_makeAnyLocalAddress}},
	{ "java/net/InetAddress.getInetFamily",{(fvoid)java_net_InetAdress_getInetFamily}},
	{ "java/net/InetAddress.getLocalHostName",{(fvoid)java_net_InetAdress_getLocalHostName}},
	{ "java/net/InetAddress.lookupAllHostAddr",{(fvoid)java_net_InetAdress_lookupAllHostAddr}},

	{ 0, 0}
};


void
invokeNative(Frame *f, char *name, arg_list args)
{
	struct function *ptf = Functions;
	int type;
	char *ptr;

DBG1(chat("invokeNative [%s]\n", name);)

	/* recupere le type de retour */
	ptr = name;
	while( *ptr++ != ')' );
	type = sign2type(ptr);

	ptr = name;
	while( *ptr != '(' ) ptr++;
	*ptr = 0;

	/* recherche la fonction puis exécution */
	while( ptf->u.i ){
		if( !strcmp(name, ptf->name) ){
			switch(type){
				case T_VOID:
					(*ptf->u.v)(args);
					break;
				case T_BOOLEAN:
				case T_INT: {
					int v = (*ptf->u.i)(args);
					if( !f->exception )
						push(f, v);
					break;
				}
				case T_FLOAT: {
					float v = (*ptf->u.f)(args);
					if( !f->exception )
						pushFloat(f, v);
					break;
				}
				case T_DOUBLE: {
					double v = (*ptf->u.d)(args);
					if( !f->exception )
						pushDouble(f, v);
					break;
				}
				case T_LONG: {
					vlong v = (*ptf->u.l)(args);
					if( !f->exception )
						pushLong(f, v);
					break;
				}
				case T_CLASS:
				case T_STRING:
				case T_TAB: {
					int v = (int)(*ptf->u.po)(args);
					if( !f->exception )
						push(f, v);
					break;
				}
				default:
					panic("no such return type in invokeNative (%s)\n", name);
			}
			break;
		}
		ptf++;
	}
	if( !ptf->name )
		panic("native function %s not found\n", name);
	
}
